;;; org-autofocus.el --- Mark Forster's Auto Focus system for Org  -*- lexical-binding: t; -*-

;; Copyright (C) 2020 David Owen

;; Author: David Owen <dsowen@fugue88.ws>
;; Keywords: autofocus
;; Package-Requires: (org)

;;; Commentary:

;;; Code:

(require 'calendar)
(require 'eieio)
(require 'ewoc)

(require 'org-agenda)


(defvar org-af-marked 'org-af-marked)
(put 'org-af-marked 'org-af-marked t)
(put 'org-af-marked 'display "*")


(defvar org-af-hidden 'org-af-hidden)
(put 'org-af-hidden 'org-af-hidden t)
(put 'org-af-hidden 'invisible t)


(defvar org-af-dimmed 'org-af-dimmed)
(put 'org-af-dimmed 'org-af-dimmed t)
(put 'org-af-dimmed 'face 'org-agenda-dimmed-todo-face)
(put 'org-af-dimmed 'evaporate t)

(defun org-af--dim (start end)
  (let ((o (make-overlay start end)))
    (overlay-put o 'category org-af-dimmed)))

(defun org-af--undim (start end)
  (remove-overlays start end 'org-af-dimmed t))

(defun org-af--dimmed-p (point)
  (get-char-property point 'org-af-dimmed))


(defun org-af-mark ()
  "Mark or unmark the current task."
  (interactive)
  (beginning-of-line)
  (when (get-text-property (point) 'org-af-hidden)
    (goto-char (next-single-property-change (point) 'org-af-hidden)))

  (let ((inhibit-read-only t)
        (task-point (point))
        bol marked)
    (unless (org-af--dimmed-p task-point)
      (setq marked (not (get-text-property task-point 'org-af-marked)))
      (if marked
          (add-text-properties task-point (1+ task-point)
                               '(category org-af-marked))
        (remove-text-properties task-point (1+ task-point)
                                '(category org-af-marked)))

      (save-excursion
        ;; Go to previous mark, if any
        (while (and (= (forward-line -1) 0)
                    (setq bol (line-beginning-position))
                    (get-text-property bol 'type)
                    (not (get-text-property bol 'org-af-marked))))

        (when (get-text-property bol 'org-af-marked)
          (if marked
              (org-af--dim bol (line-end-position))
            (org-af--undim bol (line-end-position))))

        (forward-line)
        (if marked
            (add-text-properties bol task-point '(category org-af-hidden))
          (remove-text-properties bol task-point '(category org-af-hidden)))))))


(defvar org-autofocus-mode-map (make-sparse-keymap))
(define-key org-autofocus-mode-map "." 'org-af-mark)


(define-minor-mode org-autofocus-mode nil nil " AF" org-autofocus-mode-map
  nil)


;; (defvar-local org-af--ewoc nil)


;; (defclass org-af--task ()
;;   ((org :initarg :org)
;;    (marked :initform nil)
;;    (passed :initform nil)))


;; (defun org-af--pp (obj)
;;   "Print the task OBJ in the current buffer at point."
;;   (with-slots (org marked passed) obj
;;     (when (or marked (not passed))
;;       (insert (if marked "*" " ")
;;               " "
;;               (if passed
;;                   (propertize org 'face 'org-agenda-dimmed-todo-face)
;;                 org)
;;               "\n"))))


;; (define-derived-mode org-autofocus-mode special-mode "Org AF" nil
;;   (let ((inhibit-read-only t)
;;         (date (calendar-current-date))
;;         tasks)
;;     (setq buffer-undo-list t)

;;     (dolist (f (org-agenda-files))
;;       (catch 'nextfile
;; 	(org-check-agenda-file f)
;; 	(setq tasks (nconc (org-agenda-get-day-entries f date :todo) tasks))))

;;     (setq org-af--ewoc (ewoc-create 'org-af--pp "header\n" "footer\n" t))
;;     (dolist (task tasks)
;;       (ewoc-enter-last org-af--ewoc (make-instance 'org-af--task :org task)))))


;; (defun org-af--mark ()
;;   "Toggle whether the task at point is marked."
;;   (interactive)
;;   (let* ((node (ewoc-locate org-af--ewoc))
;;          (obj (ewoc-data node))
;;          changed)
;;     (unless (slot-value obj 'passed)
;;       (let ((marked (not (slot-value obj 'marked))))
;;         (setf (slot-value obj 'marked) marked)
;;         (push node changed)
;;         (let ((node node))
;;           (while (and (setf node (ewoc-prev org-af--ewoc node))
;;                       (setf obj (ewoc-data node))
;;                       (not (slot-value obj 'marked)))
;;             (setf (slot-value obj 'passed) marked)
;;             (push node changed))
;;           (when node
;;             (setf (slot-value obj 'passed) marked)
;;             (push node changed))))
;;       (apply 'ewoc-invalidate org-af--ewoc changed)
;;       (ewoc-goto-node org-af--ewoc node))))


;; (defun org-af--complete ()
;;   "Complete the task at point, if it's marked and not passed over."
;;   (interactive)
;;   (let* ((node (ewoc-locate org-af--ewoc))
;;          (obj (ewoc-data node)))
;;     (unless (and nil (or (slot-value obj 'passed)
;;                 (not (slot-value obj 'marked))))
;;       (org-agenda-todo)
;;       (org-autofocus-mode))))


;; (define-key org-autofocus-mode-map "." 'org-af--mark)
;; (define-key org-autofocus-mode-map "c" 'org-af--complete)


;; ;;;###autoload
;; (defun org-autofocus ()
;;   "Show the AutoFocus buffer; create it if necessary."
;;   (interactive)
;;   (pop-to-buffer (get-buffer-create "*Org AutoFocus*"))
;;   (org-autofocus-mode))


(provide 'org-autofocus)
;;; org-autofocus.el ends here
